#
# Monitoring Suite, Seven Solutions S.L.
#
# This file is part of monitoring-7s.
# You might use, distribute and modify this code under the terms of
# the LICENSE.txt provided within the project monitoring-7s.
#
# If you do not have received a copy of the LICENSE.txt along with
# this file please write to info@sevensols.com and consider that
# this file can not be copied and/or distributed in any forms.
#
# Created by Eva Lopez Ruiz in October 2021.
#


from pysnmp.hlapi import *
import scapy.all as scapy
from wrz_mon.configuration import Configuration
from wrz_mon.configuration import DEVICE_TYPE_F, VERSION_F, SEC_NAME_F
from wrz_mon.configuration import AUTH_PASSWORD_F, PRIV_PASSWORD_F
from wrz_mon.configuration import IP_F, PROTOCOL_F, PORT_F
from wrz_mon.utils import LOGGER, yml_load_file, yml_dump_file


DEV_TYPES = ['wrz', 'wrs', 'ptp4l']
WRZ, WRS, PTP4L = DEV_TYPES
OID_F = 'oid'
CREDENTIALS_F = 'credentials'
OID_FILE = 'wrz_mon/autodiscovery/oids.yml'

CONFIG = Configuration()


class AutoDiscovery():
    """
    """
    def __init__(self):
        """Initialize an AutoDiscovery object"""
        self.devices = None
        self._load_configuration()
        self.arp_scan = ARPScan(CONFIG.network)
        self.snmp_scan = SNMPScan(self.devices)

    def run(self):
        """Discover devices in a network using SNMP"""
        if self.devices is None:
            LOGGER.error('OIDs not loaded')
            return

        LOGGER.info('Network to scan: %s' % CONFIG.network)

        dev_found = self.arp_scan.run()
        if dev_found == []:
            LOGGER.info('No devices found to scan')
            return

        LOGGER.info('%s devices found to scan' % len(dev_found))
        dev_found = self.snmp_scan.run(dev_found)

        if dev_found == []:
            LOGGER.info('No devices found using SNMP scan')
            return

        self._generate_configuration(dev_found)

    def _load_configuration(self):
        """
        """
        self.devices, err = yml_load_file(OID_FILE)
        if self.devices is None:
            LOGGER.error('Unable to load OIDs. Exception: %s' % err)
            return

        for dev in CONFIG.devices.values():
            if dev[DEVICE_TYPE_F] not in DEV_TYPES:
                LOGGER.warning('Unknown device type: %s' % dev[DEVICE_TYPE_F])
                continue

            dtype = dev[DEVICE_TYPE_F]
            if dev[VERSION_F] == 3:
                self.devices[dtype][CREDENTIALS_F] = UsmUserData(dev[SEC_NAME_F],
                                                                 authKey=dev[AUTH_PASSWORD_F],
                                                                 privKey=dev[PRIV_PASSWORD_F],
                                                                 authProtocol=usmHMACMD5AuthProtocol,
                                                                 privProtocol=usmDESPrivProtocol)
            else:
                self.devices[dtype][CREDENTIALS_F] = CommunityData('public')

    def _generate_configuration(self, dev_found):
        """Create new configuration file"""
        for g in CONFIG.devices:
            gtype = CONFIG.devices[g][DEVICE_TYPE_F]
            if gtype in dev_found.keys():
                for dev in dev_found[gtype]:
                    CONFIG.add_agent(g, dev[0], {IP_F: dev[1], PORT_F: 161, PROTOCOL_F: 'udp'})
        CONFIG.write()


class SNMPScan():
    """
    """
    def __init__(self, snmp_conf):
        """Initialize scan object"""
        self.snmp_conf = snmp_conf
        self.engine = SnmpEngine()
        self.context = ContextData()

    def run(self, devices):
        """Scan the network to find new devices"""
        result = {WRZ: [], WRS: [], PTP4L: []}

        for d in devices:
            for conf in self.snmp_conf:
                if (OID_F not in self.snmp_conf[conf].keys()
                        or CREDENTIALS_F not in self.snmp_conf[conf].keys()):
                    continue

                dev = self._get(str(d), self.snmp_conf[conf][OID_F], self.snmp_conf[conf][CREDENTIALS_F])
                if dev is not None and dev != '':
                    LOGGER.info('{0} device found with IP {1}'.format(conf, d))
                    result[conf].append([dev, d])
                    break

        return result

    def _get(self, target, oid, credentials, port=161):
        transport_target = UdpTransportTarget((target, port))
        object_oid = ObjectType(ObjectIdentity(oid))
        handler = getCmd(self.engine, credentials, transport_target,
                         self.context, object_oid)
        errorIndication, errorStatus, errorIndex, varBinds = next(handler)
        if varBinds == []:
            return None
        return str(varBinds[0][1])


class ARPScan():
    """
    """
    def __init__(self, network):
        """Initialize scan object"""
        self.network = network

    def run(self):
        """Scan the network and find all devices connected"""
        arp_req_frame = scapy.ARP(pdst=self.network)
        broadcast_ether_frame = scapy.Ether(dst="ff:ff:ff:ff:ff:ff")
        bro_ether_arp_req = broadcast_ether_frame / arp_req_frame
        answered_list = scapy.srp(bro_ether_arp_req, timeout=2, verbose=False)[0]

        result = []
        for i in range(0, len(answered_list)):
            result.append(answered_list[i][1].psrc)

        return result
